@@ -1,26 +1,38 @@
 # agent runner + agent worker
 from llama_index.core.agent.custom.pipeline_worker import QueryPipelineAgentWorker
 from llama_index.core.agent.custom.simple import CustomSimpleAgentWorker
+from llama_index.core.agent.custom.simple_function import FnAgentWorker
 from llama_index.core.agent.react.base import ReActAgent
 from llama_index.core.agent.react.formatter import ReActChatFormatter
+from llama_index.core.agent.react.output_parser import ReActOutputParser
 from llama_index.core.agent.react.step import ReActAgentWorker
 from llama_index.core.agent.react_multimodal.step import MultimodalReActAgentWorker
 from llama_index.core.agent.runner.base import AgentRunner
+from llama_index.core.agent.runner.planner import StructuredPlannerAgent
 from llama_index.core.agent.runner.parallel import ParallelAgentRunner
 from llama_index.core.agent.types import Task
 from llama_index.core.chat_engine.types import AgentChatResponse
+from llama_index.core.agent.function_calling.base import FunctionCallingAgent
+from llama_index.core.agent.function_calling.step import FunctionCallingAgentWorker
 
 __all__ = [
     "AgentRunner",
+    "StructuredPlannerAgent",
     "ParallelAgentRunner",
     "ReActAgentWorker",
     "ReActAgent",
+    "ReActOutputParser",
     "CustomSimpleAgentWorker",
     "QueryPipelineAgentWorker",
     "ReActChatFormatter",
+    "FunctionCallingAgentWorker",
+    "FnAgentWorker",
+    "FunctionCallingAgent",
     # beta
     "MultimodalReActAgentWorker",
     # schema-related
     "AgentChatResponse",
     "Task",
+    "TaskStep",
+    "TaskStepOutput",
 ]
