@@ -1,5 +1,6 @@
 from llama_index.core.agent.react.base import ReActAgent
 from llama_index.core.agent.react.formatter import ReActChatFormatter
 from llama_index.core.agent.react.step import ReActAgentWorker
+from llama_index.core.agent.react.output_parser import ReActOutputParser
 
-__all__ = ["ReActChatFormatter", "ReActAgentWorker", "ReActAgent"]
+__all__ = ["ReActChatFormatter", "ReActAgentWorker", "ReActAgent", "ReActOutputParser"]
