@@ -2,6 +2,7 @@ from .base import CallbackManager
 from .llama_debug import LlamaDebugHandler
 from .schema import CBEvent, CBEventType, EventPayload
 from .token_counting import TokenCountingHandler
+from .pythonically_printing_base_handler import PythonicallyPrintingBaseHandler
 from .utils import trace_method
 
 __all__ = [
@@ -12,4 +13,5 @@ __all__ = [
     "LlamaDebugHandler",
     "TokenCountingHandler",
     "trace_method",
+    "PythonicallyPrintingBaseHandler",
 ]
