@@ -12,13 +12,7 @@ from llama_index.core.data_structs.data_structs import IndexStruct
 from llama_index.core.ingestion import run_transformations
 from llama_index.core.llms.utils import LLMType, resolve_llm
 from llama_index.core.schema import BaseNode, Document, IndexNode, TransformComponent
-from llama_index.core.service_context import ServiceContext
-from llama_index.core.settings import (
-    Settings,
-    callback_manager_from_settings_or_context,
-    llm_from_settings_or_context,
-    transformations_from_settings_or_context,
-)
+from llama_index.core.settings import Settings
 from llama_index.core.storage.docstore.types import BaseDocumentStore, RefDocInfo
 from llama_index.core.storage.storage_context import StorageContext
 
@@ -34,9 +28,6 @@ class BaseIndex(Generic[IS], ABC):
     Args:
         nodes (List[Node]): List of nodes to index
         show_progress (bool): Whether to show tqdm progress bars. Defaults to False.
-        service_context (ServiceContext): Service context container (contains
-            components like LLM, Embeddings, etc.).
-
     """
 
     index_struct_cls: Type[IS]
@@ -50,14 +41,12 @@ class BaseIndex(Generic[IS], ABC):
         callback_manager: Optional[CallbackManager] = None,
         transformations: Optional[List[TransformComponent]] = None,
         show_progress: bool = False,
-        # deprecated
-        service_context: Optional[ServiceContext] = None,
         **kwargs: Any,
     ) -> None:
         """Initialize with parameters."""
         if index_struct is None and nodes is None and objects is None:
             raise ValueError("One of nodes, objects, or index_struct must be provided.")
-        if index_struct is not None and nodes is not None:
+        if index_struct is not None and nodes is not None and len(nodes) >= 1:
             raise ValueError("Only one of nodes or index_struct can be provided.")
         # This is to explicitly make sure that the old UX is not used
         if nodes is not None and len(nodes) >= 1 and not isinstance(nodes[0], BaseNode):
@@ -71,33 +60,28 @@ class BaseIndex(Generic[IS], ABC):
                 raise ValueError("nodes must be a list of Node objects.")
 
         self._storage_context = storage_context or StorageContext.from_defaults()
-        # deprecated
-        self._service_context = service_context
-
         self._docstore = self._storage_context.docstore
         self._show_progress = show_progress
         self._vector_store = self._storage_context.vector_store
         self._graph_store = self._storage_context.graph_store
-        self._callback_manager = (
-            callback_manager
-            or callback_manager_from_settings_or_context(Settings, service_context)
-        )
+        self._callback_manager = callback_manager or Settings.callback_manager
 
         objects = objects or []
         self._object_map = {obj.index_id: obj.obj for obj in objects}
+        for obj in objects:
+            obj.obj = None  # clear the object to avoid serialization issues
+
         with self._callback_manager.as_trace("index_construction"):
             if index_struct is None:
                 nodes = nodes or []
                 index_struct = self.build_index_from_nodes(
-                    nodes + objects  # type: ignore
+                    nodes + objects,  # type: ignore
+                    **kwargs,  # type: ignore
                 )
             self._index_struct = index_struct
             self._storage_context.index_store.add_index_struct(self._index_struct)
 
-        self._transformations = (
-            transformations
-            or transformations_from_settings_or_context(Settings, service_context)
-        )
+        self._transformations = transformations or Settings.transformations
 
     @classmethod
     def from_documents(
@@ -107,8 +91,6 @@ class BaseIndex(Generic[IS], ABC):
         show_progress: bool = False,
         callback_manager: Optional[CallbackManager] = None,
         transformations: Optional[List[TransformComponent]] = None,
-        # deprecated
-        service_context: Optional[ServiceContext] = None,
         **kwargs: Any,
     ) -> IndexType:
         """Create index from documents.
@@ -120,13 +102,8 @@ class BaseIndex(Generic[IS], ABC):
         """
         storage_context = storage_context or StorageContext.from_defaults()
         docstore = storage_context.docstore
-        callback_manager = (
-            callback_manager
-            or callback_manager_from_settings_or_context(Settings, service_context)
-        )
-        transformations = transformations or transformations_from_settings_or_context(
-            Settings, service_context
-        )
+        callback_manager = callback_manager or Settings.callback_manager
+        transformations = transformations or Settings.transformations
 
         with callback_manager.as_trace("index_construction"):
             for doc in documents:
@@ -145,7 +122,6 @@ class BaseIndex(Generic[IS], ABC):
                 callback_manager=callback_manager,
                 show_progress=show_progress,
                 transformations=transformations,
-                service_context=service_context,
                 **kwargs,
             )
 
@@ -182,10 +158,6 @@ class BaseIndex(Generic[IS], ABC):
         """Get the docstore corresponding to the index."""
         return self._docstore
 
-    @property
-    def service_context(self) -> Optional[ServiceContext]:
-        return self._service_context
-
     @property
     def storage_context(self) -> StorageContext:
         return self._storage_context
@@ -200,13 +172,17 @@ class BaseIndex(Generic[IS], ABC):
         self._storage_context.index_store.add_index_struct(self._index_struct)
 
     @abstractmethod
-    def _build_index_from_nodes(self, nodes: Sequence[BaseNode]) -> IS:
+    def _build_index_from_nodes(
+        self, nodes: Sequence[BaseNode], **build_kwargs: Any
+    ) -> IS:
         """Build the index from nodes."""
 
-    def build_index_from_nodes(self, nodes: Sequence[BaseNode]) -> IS:
+    def build_index_from_nodes(
+        self, nodes: Sequence[BaseNode], **build_kwargs: Any
+    ) -> IS:
         """Build the index from nodes."""
         self._docstore.add_documents(nodes, allow_update=True)
-        return self._build_index_from_nodes(nodes)
+        return self._build_index_from_nodes(nodes, **build_kwargs)
 
     @abstractmethod
     def _insert(self, nodes: Sequence[BaseNode], **insert_kwargs: Any) -> None:
@@ -214,6 +190,14 @@ class BaseIndex(Generic[IS], ABC):
 
     def insert_nodes(self, nodes: Sequence[BaseNode], **insert_kwargs: Any) -> None:
         """Insert nodes."""
+        for node in nodes:
+            if isinstance(node, IndexNode):
+                try:
+                    node.dict()
+                except ValueError:
+                    self._object_map[node.index_id] = node.obj
+                    node.obj = None
+
         with self._callback_manager.as_trace("insert_nodes"):
             self.docstore.add_documents(nodes, allow_update=True)
             self._insert(nodes, **insert_kwargs)
@@ -226,6 +210,7 @@ class BaseIndex(Generic[IS], ABC):
                 [document],
                 self._transformations,
                 show_progress=self._show_progress,
+                **insert_kwargs,
             )
 
             self.insert_nodes(nodes, **insert_kwargs)
@@ -376,6 +361,11 @@ class BaseIndex(Generic[IS], ABC):
     def as_query_engine(
         self, llm: Optional[LLMType] = None, **kwargs: Any
     ) -> BaseQueryEngine:
+        """Convert the index to a query engine.
+
+        Calls `index.as_retriever(**kwargs)` to get the retriever and then wraps it in a
+        `RetrieverQueryEngine.from_args(retriever, **kwrags)` call.
+        """
         # NOTE: lazy import
         from llama_index.core.query_engine.retriever_query_engine import (
             RetrieverQueryEngine,
@@ -385,7 +375,7 @@ class BaseIndex(Generic[IS], ABC):
         llm = (
             resolve_llm(llm, callback_manager=self._callback_manager)
             if llm
-            else llm_from_settings_or_context(Settings, self.service_context)
+            else Settings.llm
         )
 
         return RetrieverQueryEngine.from_args(
@@ -400,20 +390,25 @@ class BaseIndex(Generic[IS], ABC):
         llm: Optional[LLMType] = None,
         **kwargs: Any,
     ) -> BaseChatEngine:
-        service_context = kwargs.get("service_context", self.service_context)
+        """Convert the index to a chat engine.
 
-        if service_context is not None:
-            llm = (
-                resolve_llm(llm, callback_manager=self._callback_manager)
-                if llm
-                else service_context.llm
-            )
-        else:
-            llm = (
-                resolve_llm(llm, callback_manager=self._callback_manager)
-                if llm
-                else Settings.llm
-            )
+        Calls `index.as_query_engine(llm=llm, **kwargs)` to get the query engine and then
+        wraps it in a chat engine based on the chat mode.
+
+        Chat modes:
+            - `ChatMode.BEST` (default): Chat engine that uses an agent (react or openai) with a query engine tool
+            - `ChatMode.CONTEXT`: Chat engine that uses a retriever to get context
+            - `ChatMode.CONDENSE_QUESTION`: Chat engine that condenses questions
+            - `ChatMode.CONDENSE_PLUS_CONTEXT`: Chat engine that condenses questions and uses a retriever to get context
+            - `ChatMode.SIMPLE`: Simple chat engine that uses the LLM directly
+            - `ChatMode.REACT`: Chat engine that uses a react agent with a query engine tool
+            - `ChatMode.OPENAI`: Chat engine that uses an openai agent with a query engine tool
+        """
+        llm = (
+            resolve_llm(llm, callback_manager=self._callback_manager)
+            if llm
+            else Settings.llm
+        )
 
         query_engine = self.as_query_engine(llm=llm, **kwargs)
 
