@@ -1,10 +1,17 @@
 """Default query for SQLStructStoreIndex."""
+
 import logging
 from abc import abstractmethod
 from typing import Any, Dict, List, Optional, Tuple, Union, cast
-
+from llama_index.core.indices.base import BaseRetriever
 from llama_index.core.base.base_query_engine import BaseQueryEngine
-from llama_index.core.base.response.schema import Response
+from llama_index.core.base.response.schema import (
+    RESPONSE_TYPE,
+    AsyncStreamingResponse,
+    Response,
+    StreamingResponse,
+)
+from llama_index.core.callbacks import CallbackManager
 from llama_index.core.indices.struct_store.container_builder import (
     SQLContextContainerBuilder,
 )
@@ -16,23 +23,20 @@ from llama_index.core.indices.struct_store.sql_retriever import (
 from llama_index.core.llms.llm import LLM
 from llama_index.core.objects.base import ObjectRetriever
 from llama_index.core.objects.table_node_mapping import SQLTableSchema
+from llama_index.core.base.embeddings.base import BaseEmbedding
 from llama_index.core.prompts import BasePromptTemplate, PromptTemplate
 from llama_index.core.prompts.default_prompts import (
     DEFAULT_TEXT_TO_SQL_PGVECTOR_PROMPT,
     DEFAULT_TEXT_TO_SQL_PROMPT,
+    DEFAULT_REFINE_PROMPT,
 )
 from llama_index.core.prompts.mixin import PromptDictType, PromptMixinType
 from llama_index.core.prompts.prompt_type import PromptType
 from llama_index.core.response_synthesizers import (
     get_response_synthesizer,
 )
-from llama_index.core.schema import QueryBundle
-from llama_index.core.service_context import ServiceContext
-from llama_index.core.settings import (
-    Settings,
-    callback_manager_from_settings_or_context,
-    llm_from_settings_or_context,
-)
+from llama_index.core.schema import NodeWithScore, QueryBundle
+from llama_index.core.settings import Settings
 from llama_index.core.utilities.sql_wrapper import SQLDatabase
 from sqlalchemy import Table
 
@@ -72,8 +76,14 @@ class SQLStructStoreQueryEngine(BaseQueryEngine):
     NOTE: deprecated in favor of SQLTableRetriever, kept for backward compatibility.
 
     Runs raw SQL over a SQLStructStoreIndex. No LLM calls are made here.
+
     NOTE: this query cannot work with composed indices - if the index
     contains subindices, those subindices will not be queried.
+
+    NOTE: Any Text-to-SQL application should be aware that executing
+    arbitrary SQL queries can be a security risk. It is recommended to
+    take precautions as needed, such as using restricted roles, read-only
+    databases, sandboxing, etc.
     """
 
     def __init__(
@@ -89,11 +99,7 @@ class SQLStructStoreQueryEngine(BaseQueryEngine):
             sql_context_container or index.sql_context_container
         )
         self._sql_only = sql_only
-        super().__init__(
-            callback_manager=callback_manager_from_settings_or_context(
-                Settings, index.service_context
-            )
-        )
+        super().__init__(callback_manager=Settings.callback_manager)
 
     def _get_prompt_modules(self) -> PromptMixinType:
         """Get prompt modules."""
@@ -135,6 +141,11 @@ class NLStructStoreQueryEngine(BaseQueryEngine):
     NOTE: this query cannot work with composed indices - if the index
     contains subindices, those subindices will not be queried.
 
+    NOTE: Any Text-to-SQL application should be aware that executing
+    arbitrary SQL queries can be a security risk. It is recommended to
+    take precautions as needed, such as using restricted roles, read-only
+    databases, sandboxing, etc.
+
     Args:
         index (SQLStructStoreIndex): A SQL Struct Store Index
         text_to_sql_prompt (Optional[BasePromptTemplate]): A Text to SQL
@@ -163,10 +174,9 @@ class NLStructStoreQueryEngine(BaseQueryEngine):
     ) -> None:
         """Initialize params."""
         self._index = index
-        self._llm = llm_from_settings_or_context(Settings, index.service_context)
+        self._llm = Settings.llm
         self._sql_database = index.sql_database
         self._sql_context_container = index.sql_context_container
-        self._service_context = index.service_context
         self._ref_doc_id_column = index.ref_doc_id_column
 
         self._text_to_sql_prompt = text_to_sql_prompt or DEFAULT_TEXT_TO_SQL_PROMPT
@@ -176,16 +186,7 @@ class NLStructStoreQueryEngine(BaseQueryEngine):
         self._context_query_kwargs = context_query_kwargs or {}
         self._synthesize_response = synthesize_response
         self._sql_only = sql_only
-        super().__init__(
-            callback_manager=callback_manager_from_settings_or_context(
-                Settings, index.service_context
-            )
-        )
-
-    @property
-    def service_context(self) -> Optional[ServiceContext]:
-        """Get service context."""
-        return self._service_context
+        super().__init__(callback_manager=Settings.callback_manager)
 
     def _get_prompt_modules(self) -> PromptMixinType:
         """Get prompt modules."""
@@ -197,6 +198,23 @@ class NLStructStoreQueryEngine(BaseQueryEngine):
         sql_result_start = response.find("SQLResult:")
         if sql_result_start != -1:
             response = response[:sql_result_start]
+        # If LLM returns SQLQuery: or ```sql, extract the SQL query
+        sql_query_start = response.find("SQLQuery:")
+        if sql_query_start != -1:
+            response = response[sql_query_start:]
+            response = response.replace("SQLQuery:", "")
+        sql_markdown_start = response.find("```sql")
+        if sql_markdown_start != -1:
+            response = response.replace("```sql", "")
+        response = response.replace("```", "")
+        # If LLM talks between the end of query and SQL result
+        # find semi-colon and remove everything after it
+        semi_colon = response.find(";")
+        if semi_colon != -1:
+            response = response[: semi_colon + 1]
+        # Replace escaped single quotes, happens when
+        # there's a ' in the value (e.g. "I'm")
+        response = response.replace("\\'", "''")
         return response.strip()
 
     def _get_table_context(self, query_bundle: QueryBundle) -> str:
@@ -285,45 +303,60 @@ class NLStructStoreQueryEngine(BaseQueryEngine):
         return Response(response=response_str, metadata=metadata)
 
 
-def _validate_prompt(response_synthesis_prompt: BasePromptTemplate) -> None:
+def _validate_prompt(
+    custom_prompt: BasePromptTemplate,
+    default_prompt: BasePromptTemplate,
+) -> None:
     """Validate prompt."""
-    if (
-        response_synthesis_prompt.template_vars
-        != DEFAULT_RESPONSE_SYNTHESIS_PROMPT_V2.template_vars
-    ):
+    if custom_prompt.template_vars != default_prompt.template_vars:
         raise ValueError(
-            "response_synthesis_prompt must have the following template variables: "
-            "query_str, sql_query, context_str"
+            "custom_prompt must have the following template variables: "
+            f"{default_prompt.template_vars}"
         )
 
 
 class BaseSQLTableQueryEngine(BaseQueryEngine):
+    """Base SQL Table query engine.
+
+    NOTE: Any Text-to-SQL application should be aware that executing
+    arbitrary SQL queries can be a security risk. It is recommended to
+    take precautions as needed, such as using restricted roles, read-only
+    databases, sandboxing, etc.
+    """
+
     def __init__(
         self,
         llm: Optional[LLM] = None,
         synthesize_response: bool = True,
+        markdown_response: bool = False,
         response_synthesis_prompt: Optional[BasePromptTemplate] = None,
+        callback_manager: Optional[CallbackManager] = None,
+        refine_synthesis_prompt: Optional[BasePromptTemplate] = None,
         verbose: bool = False,
-        # deprecated
-        service_context: Optional[ServiceContext] = None,
+        streaming: bool = False,
         **kwargs: Any,
     ) -> None:
         """Initialize params."""
-        self._service_context = service_context
-        self._llm = llm or llm_from_settings_or_context(Settings, service_context)
+        self._llm = llm or Settings.llm
+        if callback_manager is not None:
+            self._llm.callback_manager = callback_manager
+
         self._response_synthesis_prompt = (
             response_synthesis_prompt or DEFAULT_RESPONSE_SYNTHESIS_PROMPT_V2
         )
+        self._refine_synthesis_prompt = refine_synthesis_prompt or DEFAULT_REFINE_PROMPT
+
         # do some basic prompt validation
-        _validate_prompt(self._response_synthesis_prompt)
-        self._synthesize_response = synthesize_response
-        self._verbose = verbose
-        super().__init__(
-            callback_manager=callback_manager_from_settings_or_context(
-                Settings, service_context
-            ),
-            **kwargs,
+        _validate_prompt(
+            self._response_synthesis_prompt, DEFAULT_RESPONSE_SYNTHESIS_PROMPT_V2
         )
+        _validate_prompt(self._refine_synthesis_prompt, DEFAULT_REFINE_PROMPT)
+
+        self._synthesize_response = synthesize_response
+        self._markdown_response = markdown_response
+        self._verbose = verbose
+        self._streaming = streaming
+        super().__init__(callback_manager=callback_manager or Settings.callback_manager)
 
     def _get_prompts(self) -> Dict[str, Any]:
         """Get prompts."""
@@ -343,12 +376,28 @@ class BaseSQLTableQueryEngine(BaseQueryEngine):
     def sql_retriever(self) -> NLSQLRetriever:
         """Get SQL retriever."""
 
-    @property
-    def service_context(self) -> Optional[ServiceContext]:
-        """Get service context."""
-        return self._service_context
+    def _format_result_markdown(self, retrieved_nodes: List[NodeWithScore]) -> str:
+        """Format the result in markdown."""
+        tables = []
+        for node_with_score in retrieved_nodes:
+            node = node_with_score.node
+            metadata = node.metadata
 
-    def _query(self, query_bundle: QueryBundle) -> Response:
+            col_keys = metadata.get("col_keys", [])
+            results = metadata.get("result", [])
+            table_header = "| " + " | ".join(col_keys) + " |\n"
+            table_separator = "|" + "|".join(["---"] * len(col_keys)) + "|\n"
+
+            table_rows = ""
+            for row in results:
+                table_rows += "| " + " | ".join(str(item) for item in row) + " |\n"
+
+            markdown_table = table_header + table_separator + table_rows
+            tables.append(markdown_table)
+
+        return "\n\n".join(tables).strip()
+
+    def _query(self, query_bundle: QueryBundle) -> RESPONSE_TYPE:
         """Answer a query."""
         retrieved_nodes, metadata = self.sql_retriever.retrieve_with_metadata(
             query_bundle
@@ -363,19 +412,26 @@ class BaseSQLTableQueryEngine(BaseQueryEngine):
                 llm=self._llm,
                 callback_manager=self.callback_manager,
                 text_qa_template=partial_synthesis_prompt,
+                refine_template=self._refine_synthesis_prompt,
                 verbose=self._verbose,
+                streaming=self._streaming,
             )
             response = response_synthesizer.synthesize(
                 query=query_bundle.query_str,
                 nodes=retrieved_nodes,
             )
             cast(Dict, response.metadata).update(metadata)
+            if self._streaming:
+                return cast(StreamingResponse, response)
             return cast(Response, response)
         else:
-            response_str = "\n".join([node.node.text for node in retrieved_nodes])
+            if self._markdown_response:
+                response_str = self._format_result_markdown(retrieved_nodes)
+            else:
+                response_str = "\n".join([node.text for node in retrieved_nodes])
             return Response(response=response_str, metadata=metadata)
 
-    async def _aquery(self, query_bundle: QueryBundle) -> Response:
+    async def _aquery(self, query_bundle: QueryBundle) -> RESPONSE_TYPE:
         """Answer a query."""
         retrieved_nodes, metadata = await self.sql_retriever.aretrieve_with_metadata(
             query_bundle
@@ -386,19 +442,25 @@ class BaseSQLTableQueryEngine(BaseQueryEngine):
             partial_synthesis_prompt = self._response_synthesis_prompt.partial_format(
                 sql_query=sql_query_str,
             )
+
             response_synthesizer = get_response_synthesizer(
                 llm=self._llm,
                 callback_manager=self.callback_manager,
                 text_qa_template=partial_synthesis_prompt,
+                refine_template=self._refine_synthesis_prompt,
+                verbose=self._verbose,
+                streaming=self._streaming,
             )
             response = await response_synthesizer.asynthesize(
                 query=query_bundle.query_str,
                 nodes=retrieved_nodes,
             )
             cast(Dict, response.metadata).update(metadata)
+            if self._streaming:
+                return cast(AsyncStreamingResponse, response)
             return cast(Response, response)
         else:
-            response_str = "\n".join([node.node.text for node in retrieved_nodes])
+            response_str = "\n".join([node.text for node in retrieved_nodes])
             return Response(response=response_str, metadata=metadata)
 
 
@@ -407,6 +469,11 @@ class NLSQLTableQueryEngine(BaseSQLTableQueryEngine):
     Natural language SQL Table query engine.
 
     Read NLStructStoreQueryEngine's docstring for more info on NL SQL.
+
+    NOTE: Any Text-to-SQL application should be aware that executing
+    arbitrary SQL queries can be a security risk. It is recommended to
+    take precautions as needed, such as using restricted roles, read-only
+    databases, sandboxing, etc.
     """
 
     def __init__(
@@ -416,11 +483,14 @@ class NLSQLTableQueryEngine(BaseSQLTableQueryEngine):
         text_to_sql_prompt: Optional[BasePromptTemplate] = None,
         context_query_kwargs: Optional[dict] = None,
         synthesize_response: bool = True,
+        markdown_response: bool = False,
         response_synthesis_prompt: Optional[BasePromptTemplate] = None,
+        refine_synthesis_prompt: Optional[BasePromptTemplate] = None,
         tables: Optional[Union[List[str], List[Table]]] = None,
-        service_context: Optional[ServiceContext] = None,
         context_str_prefix: Optional[str] = None,
+        embed_model: Optional[BaseEmbedding] = None,
         sql_only: bool = False,
+        callback_manager: Optional[CallbackManager] = None,
         verbose: bool = False,
         **kwargs: Any,
     ) -> None:
@@ -433,15 +503,18 @@ class NLSQLTableQueryEngine(BaseSQLTableQueryEngine):
             context_query_kwargs=context_query_kwargs,
             tables=tables,
             context_str_prefix=context_str_prefix,
-            service_context=service_context,
+            embed_model=embed_model,
             sql_only=sql_only,
+            callback_manager=callback_manager,
             verbose=verbose,
         )
         super().__init__(
             synthesize_response=synthesize_response,
+            markdown_response=markdown_response,
             response_synthesis_prompt=response_synthesis_prompt,
+            refine_synthesis_prompt=refine_synthesis_prompt,
             llm=llm,
-            service_context=service_context,
+            callback_manager=callback_manager,
             verbose=verbose,
             **kwargs,
         )
@@ -460,6 +533,11 @@ class PGVectorSQLQueryEngine(BaseSQLTableQueryEngine):
 
     NOTE: this is a beta feature
 
+    NOTE: Any Text-to-SQL application should be aware that executing
+    arbitrary SQL queries can be a security risk. It is recommended to
+    take precautions as needed, such as using restricted roles, read-only
+    databases, sandboxing, etc.
+
     """
 
     def __init__(
@@ -470,10 +548,11 @@ class PGVectorSQLQueryEngine(BaseSQLTableQueryEngine):
         context_query_kwargs: Optional[dict] = None,
         synthesize_response: bool = True,
         response_synthesis_prompt: Optional[BasePromptTemplate] = None,
+        refine_synthesis_prompt: Optional[BasePromptTemplate] = None,
         tables: Optional[Union[List[str], List[Table]]] = None,
-        service_context: Optional[ServiceContext] = None,
         context_str_prefix: Optional[str] = None,
         sql_only: bool = False,
+        callback_manager: Optional[CallbackManager] = None,
         **kwargs: Any,
     ) -> None:
         """Initialize params."""
@@ -486,14 +565,15 @@ class PGVectorSQLQueryEngine(BaseSQLTableQueryEngine):
             tables=tables,
             sql_parser_mode=SQLParserMode.PGVECTOR,
             context_str_prefix=context_str_prefix,
-            service_context=service_context,
             sql_only=sql_only,
+            callback_manager=callback_manager,
         )
         super().__init__(
             synthesize_response=synthesize_response,
             response_synthesis_prompt=response_synthesis_prompt,
+            refine_synthesis_prompt=refine_synthesis_prompt,
             llm=llm,
-            service_context=service_context,
+            callback_manager=callback_manager,
             **kwargs,
         )
 
@@ -510,14 +590,16 @@ class SQLTableRetrieverQueryEngine(BaseSQLTableQueryEngine):
         self,
         sql_database: SQLDatabase,
         table_retriever: ObjectRetriever[SQLTableSchema],
+        rows_retrievers: Optional[dict[str, BaseRetriever]] = None,
         llm: Optional[LLM] = None,
         text_to_sql_prompt: Optional[BasePromptTemplate] = None,
         context_query_kwargs: Optional[dict] = None,
         synthesize_response: bool = True,
         response_synthesis_prompt: Optional[BasePromptTemplate] = None,
-        service_context: Optional[ServiceContext] = None,
+        refine_synthesis_prompt: Optional[BasePromptTemplate] = None,
         context_str_prefix: Optional[str] = None,
         sql_only: bool = False,
+        callback_manager: Optional[CallbackManager] = None,
         **kwargs: Any,
     ) -> None:
         """Initialize params."""
@@ -527,15 +609,18 @@ class SQLTableRetrieverQueryEngine(BaseSQLTableQueryEngine):
             text_to_sql_prompt=text_to_sql_prompt,
             context_query_kwargs=context_query_kwargs,
             table_retriever=table_retriever,
+            rows_retrievers=rows_retrievers,
             context_str_prefix=context_str_prefix,
-            service_context=service_context,
             sql_only=sql_only,
+            callback_manager=callback_manager,
+            verbose=kwargs.get("verbose", False),
         )
         super().__init__(
             synthesize_response=synthesize_response,
             response_synthesis_prompt=response_synthesis_prompt,
+            refine_synthesis_prompt=refine_synthesis_prompt,
             llm=llm,
-            service_context=service_context,
+            callback_manager=callback_manager,
             **kwargs,
         )
 
