@@ -13,8 +13,14 @@ from typing import (
     Tuple,
     Union,
 )
+from typing_extensions import Annotated
 
-from llama_index.core.bridge.pydantic import Field
+from llama_index.core.bridge.pydantic import (
+    Field,
+    WithJsonSchema,
+    PlainSerializer,
+    SerializeAsAny,
+)
 
 if TYPE_CHECKING:
     from llama_index.core.bridge.langchain import (
@@ -33,7 +39,7 @@ from llama_index.core.base.query_pipeline.query import (
     QueryComponent,
     validate_and_convert_stringable,
 )
-from llama_index.core.bridge.pydantic import BaseModel
+from llama_index.core.bridge.pydantic import BaseModel, ConfigDict
 from llama_index.core.base.llms.base import BaseLLM
 from llama_index.core.base.llms.generic_utils import (
     messages_to_prompt as default_messages_to_prompt,
@@ -41,21 +47,32 @@ from llama_index.core.base.llms.generic_utils import (
 from llama_index.core.base.llms.generic_utils import (
     prompt_to_messages,
 )
+from llama_index.core.base.llms.types import ContentBlock, TextBlock
 from llama_index.core.prompts.prompt_type import PromptType
-from llama_index.core.prompts.utils import get_template_vars
+from llama_index.core.prompts.utils import get_template_vars, format_string
 from llama_index.core.types import BaseOutputParser
 
 
-class BasePromptTemplate(ChainableMixin, BaseModel, ABC):
+AnnotatedCallable = Annotated[
+    Callable,
+    WithJsonSchema({"type": "string"}),
+    WithJsonSchema({"type": "string"}),
+    PlainSerializer(lambda x: f"{x.__module__}.{x.__name__}", return_type=str),
+]
+
+
+class BasePromptTemplate(ChainableMixin, BaseModel, ABC):  # type: ignore[no-redef]
+    model_config = ConfigDict(arbitrary_types_allowed=True)
     metadata: Dict[str, Any]
     template_vars: List[str]
     kwargs: Dict[str, str]
     output_parser: Optional[BaseOutputParser]
     template_var_mappings: Optional[Dict[str, Any]] = Field(
-        default_factory=dict, description="Template variable mappings (Optional)."
+        default_factory=dict,  # type: ignore
+        description="Template variable mappings (Optional).",
     )
-    function_mappings: Optional[Dict[str, Callable]] = Field(
-        default_factory=dict,
+    function_mappings: Optional[Dict[str, AnnotatedCallable]] = Field(
+        default_factory=dict,  # type: ignore
         description=(
             "Function mappings (Optional). This is a mapping from template "
             "variable names to functions that take in the current kwargs and "
@@ -106,9 +123,6 @@ class BasePromptTemplate(ChainableMixin, BaseModel, ABC):
         # map template vars (to point to existing format vars in string template)
         return self._map_template_vars(new_kwargs)
 
-    class Config:
-        arbitrary_types_allowed = True
-
     @abstractmethod
     def partial_format(self, **kwargs: Any) -> "BasePromptTemplate":
         ...
@@ -134,7 +148,7 @@ class BasePromptTemplate(ChainableMixin, BaseModel, ABC):
         return PromptComponent(prompt=self, format_messages=False, llm=llm)
 
 
-class PromptTemplate(BasePromptTemplate):
+class PromptTemplate(BasePromptTemplate):  # type: ignore[no-redef]
     template: str
 
     def __init__(
@@ -193,7 +207,7 @@ class PromptTemplate(BasePromptTemplate):
         }
 
         mapped_all_kwargs = self._map_all_vars(all_kwargs)
-        prompt = self.template.format(**mapped_all_kwargs)
+        prompt = format_string(self.template, **mapped_all_kwargs)
 
         if self.output_parser is not None:
             prompt = self.output_parser.format(prompt)
@@ -215,12 +229,12 @@ class PromptTemplate(BasePromptTemplate):
         return self.template
 
 
-class ChatPromptTemplate(BasePromptTemplate):
+class ChatPromptTemplate(BasePromptTemplate):  # type: ignore[no-redef]
     message_templates: List[ChatMessage]
 
     def __init__(
         self,
-        message_templates: List[ChatMessage],
+        message_templates: Sequence[ChatMessage],
         prompt_type: str = PromptType.CUSTOM,
         output_parser: Optional[BaseOutputParser] = None,
         metadata: Optional[Dict[str, Any]] = None,
@@ -246,6 +260,20 @@ class ChatPromptTemplate(BasePromptTemplate):
             function_mappings=function_mappings,
         )
 
+    @classmethod
+    def from_messages(
+        cls,
+        message_templates: Union[List[Tuple[str, str]], List[ChatMessage]],
+        **kwargs: Any,
+    ) -> "ChatPromptTemplate":
+        """From messages."""
+        if isinstance(message_templates[0], tuple):
+            message_templates = [
+                ChatMessage.from_str(role=role, content=content)  # type: ignore[arg-type]
+                for role, content in message_templates
+            ]
+        return cls(message_templates=message_templates, **kwargs)  # type: ignore[arg-type]
+
     def partial_format(self, **kwargs: Any) -> "ChatPromptTemplate":
         prompt = deepcopy(self)
         prompt.kwargs.update(kwargs)
@@ -278,18 +306,30 @@ class ChatPromptTemplate(BasePromptTemplate):
 
         messages: List[ChatMessage] = []
         for message_template in self.message_templates:
-            template_vars = get_template_vars(message_template.content or "")
-            relevant_kwargs = {
-                k: v for k, v in mapped_all_kwargs.items() if k in template_vars
-            }
-            content_template = message_template.content or ""
+            # Handle messages with multiple blocks
+            if message_template.blocks:
+                formatted_blocks: List[ContentBlock] = []
+                for block in message_template.blocks:
+                    if isinstance(block, TextBlock):
+                        template_vars = get_template_vars(block.text)
+                        relevant_kwargs = {
+                            k: v
+                            for k, v in mapped_all_kwargs.items()
+                            if k in template_vars
+                        }
+                        formatted_text = format_string(block.text, **relevant_kwargs)
+                        formatted_blocks.append(TextBlock(text=formatted_text))
+                    else:
+                        # For non-text blocks (like images), keep them as is
+                        # TODO: can images be formatted as variables?
+                        formatted_blocks.append(block)
 
-            # if there's mappings specified, make sure those are used
-            content = content_template.format(**relevant_kwargs)
-
-            message: ChatMessage = message_template.copy()
-            message.content = content
-            messages.append(message)
+                message = message_template.model_copy()
+                message.blocks = formatted_blocks
+                messages.append(message)
+            else:
+                # Handle empty messages (if any)
+                messages.append(message_template.model_copy())
 
         if self.output_parser is not None:
             messages = self.output_parser.format_messages(messages)
@@ -306,17 +346,17 @@ class ChatPromptTemplate(BasePromptTemplate):
         return PromptComponent(prompt=self, format_messages=True, llm=llm)
 
 
-class SelectorPromptTemplate(BasePromptTemplate):
-    default_template: BasePromptTemplate
+class SelectorPromptTemplate(BasePromptTemplate):  # type: ignore[no-redef]
+    default_template: SerializeAsAny[BasePromptTemplate]
     conditionals: Optional[
-        List[Tuple[Callable[[BaseLLM], bool], BasePromptTemplate]]
+        Sequence[Tuple[Callable[[BaseLLM], bool], BasePromptTemplate]]
     ] = None
 
     def __init__(
         self,
         default_template: BasePromptTemplate,
         conditionals: Optional[
-            List[Tuple[Callable[[BaseLLM], bool], BasePromptTemplate]]
+            Sequence[Tuple[Callable[[BaseLLM], bool], BasePromptTemplate]]
         ] = None,
     ):
         metadata = default_template.metadata
@@ -378,7 +418,7 @@ class SelectorPromptTemplate(BasePromptTemplate):
         return prompt.get_template(llm=llm)
 
 
-class LangchainPromptTemplate(BasePromptTemplate):
+class LangchainPromptTemplate(BasePromptTemplate):  # type: ignore[no-redef]
     selector: Any
     requires_langchain_llm: bool = False
 
@@ -527,8 +567,9 @@ Prompt = PromptTemplate
 class PromptComponent(QueryComponent):
     """Prompt component."""
 
-    prompt: BasePromptTemplate = Field(..., description="Prompt")
-    llm: Optional[BaseLLM] = Field(
+    model_config = ConfigDict(arbitrary_types_allowed=True)
+    prompt: SerializeAsAny[BasePromptTemplate] = Field(..., description="Prompt")
+    llm: Optional[SerializeAsAny[BaseLLM]] = Field(
         default=None, description="LLM to use for formatting prompt."
     )
     format_messages: bool = Field(
@@ -536,9 +577,6 @@ class PromptComponent(QueryComponent):
         description="Whether to format the prompt into a list of chat messages.",
     )
 
-    class Config:
-        arbitrary_types_allowed = True
-
     def set_callback_manager(self, callback_manager: Any) -> None:
         """Set callback manager."""
 
