@@ -15,6 +15,8 @@ from llama_index.core.query_pipeline.components.router import (
     SelectorComponent,
 )
 from llama_index.core.query_pipeline.components.tool_runner import ToolRunnerComponent
+from llama_index.core.query_pipeline.components.stateful import StatefulFnComponent
+from llama_index.core.query_pipeline.components.loop import LoopComponent
 
 __all__ = [
     "AgentFnComponent",
@@ -28,4 +30,6 @@ __all__ = [
     "RouterComponent",
     "SelectorComponent",
     "ToolRunnerComponent",
+    "StatefulFnComponent",
+    "LoopComponent",
 ]
