@@ -1,23 +1,66 @@
 """Simple reader that reads files of different formats from a directory."""
 
+from __future__ import annotations
+
+import asyncio
 import logging
 import mimetypes
 import multiprocessing
+import os
 import warnings
-from datetime import datetime
+from abc import ABC, abstractmethod
+from datetime import datetime, timezone
 from functools import reduce
 from itertools import repeat
-from pathlib import Path
+from pathlib import Path, PurePosixPath
+from typing import (
+    Any,
+    Callable,
+    Generator,
+    Type,
+    cast,
+)
+
 import fsspec
 from fsspec.implementations.local import LocalFileSystem
-from typing import Any, Callable, Dict, Generator, List, Optional, Type
-
-from llama_index.core.readers.base import BaseReader
-from llama_index.core.schema import Document
 from tqdm import tqdm
 
+from llama_index.core.async_utils import get_asyncio_module, run_jobs
+from llama_index.core.readers.base import BaseReader, ResourcesReaderMixin
+from llama_index.core.schema import Document
 
-def _try_loading_included_file_formats() -> Dict[str, Type[BaseReader]]:
+
+class FileSystemReaderMixin(ABC):
+    @abstractmethod
+    def read_file_content(self, input_file: Path, **kwargs: Any) -> bytes:
+        """
+        Read the bytes content of a file.
+
+        Args:
+            input_file (Path): Path to the file.
+
+        Returns:
+            bytes: File content.
+        """
+
+    async def aread_file_content(
+        self, input_file: Path, **kwargs: Any
+    ) -> bytes:  # pragma: no cover
+        """
+        A thin wrapper around read_file_content.
+
+        Args:
+            input_file (Path): Path to the file.
+
+        Returns:
+            bytes: File content.
+        """
+        return self.read_file_content(input_file, **kwargs)
+
+
+def _try_loading_included_file_formats() -> (
+    dict[str, Type[BaseReader]]
+):  # pragma: no cover
     try:
         from llama_index.readers.file import (
             DocxReader,
@@ -25,9 +68,9 @@ def _try_loading_included_file_formats() -> Dict[str, Type[BaseReader]]:
             HWPReader,
             ImageReader,
             IPYNBReader,
-            MarkdownReader,
             MboxReader,
             PandasCSVReader,
+            PandasExcelReader,
             PDFReader,
             PptxReader,
             VideoAudioReader,
@@ -35,58 +78,84 @@ def _try_loading_included_file_formats() -> Dict[str, Type[BaseReader]]:
     except ImportError:
         raise ImportError("`llama-index-readers-file` package not found")
 
-    default_file_reader_cls: Dict[str, Type[BaseReader]] = {
+    default_file_reader_cls: dict[str, Type[BaseReader]] = {
         ".hwp": HWPReader,
         ".pdf": PDFReader,
         ".docx": DocxReader,
         ".pptx": PptxReader,
         ".ppt": PptxReader,
         ".pptm": PptxReader,
+        ".gif": ImageReader,
         ".jpg": ImageReader,
         ".png": ImageReader,
         ".jpeg": ImageReader,
+        ".webp": ImageReader,
         ".mp3": VideoAudioReader,
         ".mp4": VideoAudioReader,
         ".csv": PandasCSVReader,
         ".epub": EpubReader,
-        ".md": MarkdownReader,
         ".mbox": MboxReader,
         ".ipynb": IPYNBReader,
+        ".xls": PandasExcelReader,
+        ".xlsx": PandasExcelReader,
     }
     return default_file_reader_cls
 
 
-def _format_file_timestamp(timestamp: float) -> Optional[str]:
-    """Format file timestamp to a %Y-%m-%d string.
+def _format_file_timestamp(
+    timestamp: float | None, include_time: bool = False
+) -> str | None:
+    """
+    Format file timestamp to a string.
+    The format will be %Y-%m-%d if include_time is False or missing,
+    %Y-%m-%dT%H:%M:%SZ if include_time is True.
 
     Args:
         timestamp (float): timestamp in float
+        include_time (bool): whether to include time in the formatted string
 
     Returns:
         str: formatted timestamp
+        None: if the timestamp passed was None
     """
-    try:
-        return datetime.fromtimestamp(timestamp).strftime("%Y-%m-%d")
-    except Exception:
+    if timestamp is None:
         return None
 
+    # Convert timestamp to UTC
+    # Check if timestamp is already a datetime object
+    if isinstance(timestamp, datetime):
+        timestamp_dt = timestamp.astimezone(timezone.utc)
+    else:
+        timestamp_dt = datetime.fromtimestamp(timestamp, tz=timezone.utc)
+
+    if include_time:
+        return timestamp_dt.strftime("%Y-%m-%dT%H:%M:%SZ")
+    return timestamp_dt.strftime("%Y-%m-%d")
+
 
 def default_file_metadata_func(
-    file_path: str, fs: Optional[fsspec.AbstractFileSystem] = None
-) -> Dict:
-    """Get some handy metadate from filesystem.
+    file_path: str, fs: fsspec.AbstractFileSystem | None = None
+) -> dict:
+    """
+    Get some handy metadata from filesystem.
 
     Args:
         file_path: str: file path in str
     """
     fs = fs or get_default_fs()
     stat_result = fs.stat(file_path)
+
+    try:
+        file_name = os.path.basename(str(stat_result["name"]))
+    except Exception as e:
+        file_name = os.path.basename(file_path)
+
     creation_date = _format_file_timestamp(stat_result.get("created"))
     last_modified_date = _format_file_timestamp(stat_result.get("mtime"))
     last_accessed_date = _format_file_timestamp(stat_result.get("atime"))
     default_meta = {
         "file_path": file_path,
-        "file_name": stat_result["name"],
+        "file_name": file_name,
         "file_type": mimetypes.guess_type(file_path)[0],
         "file_size": stat_result.get("size"),
         "creation_date": creation_date,
@@ -108,10 +177,10 @@ class _DefaultFileMetadataFunc:
     Allows for pickling of the function.
     """
 
-    def __init__(self, fs: Optional[fsspec.AbstractFileSystem] = None):
+    def __init__(self, fs: fsspec.AbstractFileSystem | None = None):
         self.fs = fs or get_default_fs()
 
-    def __call__(self, file_path: str) -> Dict:
+    def __call__(self, file_path: str) -> dict:
         return default_file_metadata_func(file_path, self.fs)
 
 
@@ -126,18 +195,20 @@ def is_default_fs(fs: fsspec.AbstractFileSystem) -> bool:
 logger = logging.getLogger(__name__)
 
 
-class SimpleDirectoryReader(BaseReader):
-    """Simple directory reader.
+class SimpleDirectoryReader(BaseReader, ResourcesReaderMixin, FileSystemReaderMixin):
+    """
+    Simple directory reader.
 
     Load files from file directory.
     Automatically select the best file reader given file extensions.
 
     Args:
-        input_dir (str): Path to the directory.
+        input_dir (Union[Path, str]): Path to the directory.
         input_files (List): List of file paths to read
             (Optional; overrides input_dir, exclude)
         exclude (List): glob of python file paths to exclude (Optional)
         exclude_hidden (bool): Whether to exclude hidden files (dotfiles).
+        exclude_empty (bool): Whether to exclude empty files (Optional).
         encoding (str): Encoding of the files.
             Default is utf-8.
         errors (str): how encoding and decoding errors are to be handled,
@@ -156,6 +227,7 @@ class SimpleDirectoryReader(BaseReader):
         file_metadata (Optional[Callable[str, Dict]]): A function that takes
             in a filename and returns a Dict of metadata for the Document.
             Default is None.
+        raise_on_error (bool): Whether to raise an error if a file cannot be read.
         fs (Optional[fsspec.AbstractFileSystem]): File system to use. Defaults
         to using the local file system. Can be changed to use any remote file system
         exposed via the fsspec interface.
@@ -165,19 +237,21 @@ class SimpleDirectoryReader(BaseReader):
 
     def __init__(
         self,
-        input_dir: Optional[str] = None,
-        input_files: Optional[List] = None,
-        exclude: Optional[List] = None,
+        input_dir: Path | str | None = None,
+        input_files: list | None = None,
+        exclude: list | None = None,
         exclude_hidden: bool = True,
+        exclude_empty: bool = False,
         errors: str = "ignore",
         recursive: bool = False,
         encoding: str = "utf-8",
         filename_as_id: bool = False,
-        required_exts: Optional[List[str]] = None,
-        file_extractor: Optional[Dict[str, BaseReader]] = None,
-        num_files_limit: Optional[int] = None,
-        file_metadata: Optional[Callable[[str], Dict]] = None,
-        fs: Optional[fsspec.AbstractFileSystem] = None,
+        required_exts: list[str] | None = None,
+        file_extractor: dict[str, BaseReader] | None = None,
+        num_files_limit: int | None = None,
+        file_metadata: Callable[[str], dict] | None = None,
+        raise_on_error: bool = False,
+        fs: fsspec.AbstractFileSystem | None = None,
     ) -> None:
         """Initialize with parameters."""
         super().__init__()
@@ -192,76 +266,101 @@ class SimpleDirectoryReader(BaseReader):
         self.exclude = exclude
         self.recursive = recursive
         self.exclude_hidden = exclude_hidden
+        self.exclude_empty = exclude_empty
         self.required_exts = required_exts
         self.num_files_limit = num_files_limit
+        self.raise_on_error = raise_on_error
+        _Path = Path if is_default_fs(self.fs) else PurePosixPath
 
         if input_files:
             self.input_files = []
             for path in input_files:
                 if not self.fs.isfile(path):
                     raise ValueError(f"File {path} does not exist.")
-                input_file = Path(path)
+                input_file = _Path(path)
                 self.input_files.append(input_file)
         elif input_dir:
             if not self.fs.isdir(input_dir):
                 raise ValueError(f"Directory {input_dir} does not exist.")
-            self.input_dir = Path(input_dir)
+            self.input_dir = _Path(input_dir)
             self.exclude = exclude
             self.input_files = self._add_files(self.input_dir)
 
-        if file_extractor is not None:
-            self.file_extractor = file_extractor
-        else:
-            self.file_extractor = {}
-
+        self.file_extractor = file_extractor or {}
         self.file_metadata = file_metadata or _DefaultFileMetadataFunc(self.fs)
         self.filename_as_id = filename_as_id
 
-    def is_hidden(self, path: Path) -> bool:
+    def is_hidden(self, path: Path | PurePosixPath) -> bool:
         return any(
             part.startswith(".") and part not in [".", ".."] for part in path.parts
         )
 
-    def _add_files(self, input_dir: Path) -> List[Path]:
+    def is_empty_file(self, path: Path | PurePosixPath) -> bool:
+        if isinstance(path, PurePosixPath):
+            path = Path(path)
+        return path.is_file() and len(path.read_bytes()) == 0
+
+    def _add_files(self, input_dir: Path | PurePosixPath) -> list[Path | PurePosixPath]:
         """Add files."""
-        all_files = set()
-        rejected_files = set()
+        all_files: set[Path | PurePosixPath] = set()
+        rejected_files: set[Path | PurePosixPath] = set()
+        rejected_dirs: set[Path | PurePosixPath] = set()
+        # Default to POSIX paths for non-default file systems (e.g. S3)
+        _Path = Path if is_default_fs(self.fs) else PurePosixPath
 
         if self.exclude is not None:
             for excluded_pattern in self.exclude:
                 if self.recursive:
                     # Recursive glob
-                    for file in input_dir.rglob(excluded_pattern):
-                        rejected_files.add(Path(file))
+                    excluded_glob = _Path(input_dir) / _Path("**") / excluded_pattern
                 else:
                     # Non-recursive glob
-                    for file in input_dir.glob(excluded_pattern):
-                        rejected_files.add(Path(file))
+                    excluded_glob = _Path(input_dir) / excluded_pattern
+                for file in self.fs.glob(str(excluded_glob)):
+                    if self.fs.isdir(file):
+                        rejected_dirs.add(_Path(str(file)))
+                    else:
+                        rejected_files.add(_Path(str(file)))
 
-        file_refs: List[str] = []
+        file_refs: list[str] = []
         if self.recursive:
-            file_refs = self.fs.glob(str(input_dir) + "/**/*")
+            file_refs = cast(list[str], self.fs.glob(str(input_dir) + "/**/*"))
         else:
-            file_refs = self.fs.glob(str(input_dir) + "/*")
+            file_refs = cast(list[str], self.fs.glob(str(input_dir) + "/*"))
 
-        for ref in file_refs:
+        for _ref in file_refs:
             # Manually check if file is hidden or directory instead of
             # in glob for backwards compatibility.
-            ref = Path(ref)
+            ref = _Path(_ref)
             is_dir = self.fs.isdir(ref)
-            skip_because_hidden = self.exclude_hidden and self.is_hidden(
-                ref.relative_to(input_dir.absolute())
-            )
+            skip_because_hidden = self.exclude_hidden and self.is_hidden(ref)
+            skip_because_empty = self.exclude_empty and self.is_empty_file(ref)
             skip_because_bad_ext = (
                 self.required_exts is not None and ref.suffix not in self.required_exts
             )
             skip_because_excluded = ref in rejected_files
+            if not skip_because_excluded:
+                if is_dir:
+                    ref_parent_dir = ref
+                else:
+                    ref_parent_dir = self.fs._parent(ref)
+                for rejected_dir in rejected_dirs:
+                    if str(ref_parent_dir).startswith(str(rejected_dir)):
+                        skip_because_excluded = True
+                        logger.debug(
+                            "Skipping %s because it in parent dir %s which is in %s",
+                            ref,
+                            ref_parent_dir,
+                            rejected_dir,
+                        )
+                        break
 
             if (
                 is_dir
                 or skip_because_hidden
                 or skip_because_bad_ext
                 or skip_because_excluded
+                or skip_because_empty
             ):
                 continue
             else:
@@ -282,8 +381,9 @@ class SimpleDirectoryReader(BaseReader):
 
         return new_input_files
 
-    def _exclude_metadata(self, documents: List[Document]) -> List[Document]:
-        """Exclude metadata from documents.
+    def _exclude_metadata(self, documents: list[Document]) -> list[Document]:
+        """
+        Exclude metadata from documents.
 
         Args:
             documents (List[Document]): List of documents.
@@ -316,17 +416,102 @@ class SimpleDirectoryReader(BaseReader):
 
         return documents
 
+    def list_resources(self, *args: Any, **kwargs: Any) -> list[str]:
+        """List files in the given filesystem."""
+        return [str(x) for x in self.input_files]
+
+    def get_resource_info(self, resource_id: str, *args: Any, **kwargs: Any) -> dict:
+        info_result = self.fs.info(resource_id)
+
+        creation_date = _format_file_timestamp(
+            info_result.get("created"), include_time=True
+        )
+        last_modified_date = _format_file_timestamp(
+            info_result.get("mtime"), include_time=True
+        )
+
+        info_dict = {
+            "file_path": resource_id,
+            "file_size": info_result.get("size"),
+            "creation_date": creation_date,
+            "last_modified_date": last_modified_date,
+        }
+
+        # Ignore None values
+        return {
+            meta_key: meta_value
+            for meta_key, meta_value in info_dict.items()
+            if meta_value is not None
+        }
+
+    def load_resource(
+        self, resource_id: str, *args: Any, **kwargs: Any
+    ) -> list[Document]:
+        file_metadata = kwargs.get("file_metadata", self.file_metadata)
+        file_extractor = kwargs.get("file_extractor", self.file_extractor)
+        filename_as_id = kwargs.get("filename_as_id", self.filename_as_id)
+        encoding = kwargs.get("encoding", self.encoding)
+        errors = kwargs.get("errors", self.errors)
+        raise_on_error = kwargs.get("raise_on_error", self.raise_on_error)
+        fs = kwargs.get("fs", self.fs)
+
+        path_func = Path if is_default_fs(fs) else PurePosixPath
+
+        return SimpleDirectoryReader.load_file(
+            input_file=path_func(resource_id),
+            file_metadata=file_metadata,
+            file_extractor=file_extractor,
+            filename_as_id=filename_as_id,
+            encoding=encoding,
+            errors=errors,
+            raise_on_error=raise_on_error,
+            fs=fs,
+            **kwargs,
+        )
+
+    async def aload_resource(
+        self, resource_id: str, *args: Any, **kwargs: Any
+    ) -> list[Document]:
+        file_metadata = kwargs.get("file_metadata", self.file_metadata)
+        file_extractor = kwargs.get("file_extractor", self.file_extractor)
+        filename_as_id = kwargs.get("filename_as_id", self.filename_as_id)
+        encoding = kwargs.get("encoding", self.encoding)
+        errors = kwargs.get("errors", self.errors)
+        raise_on_error = kwargs.get("raise_on_error", self.raise_on_error)
+        fs = kwargs.get("fs", self.fs)
+
+        return await SimpleDirectoryReader.aload_file(
+            input_file=Path(resource_id),
+            file_metadata=file_metadata,
+            file_extractor=file_extractor,
+            filename_as_id=filename_as_id,
+            encoding=encoding,
+            errors=errors,
+            raise_on_error=raise_on_error,
+            fs=fs,
+            **kwargs,
+        )
+
+    def read_file_content(self, input_file: Path, **kwargs: Any) -> bytes:
+        """Read file content."""
+        fs: fsspec.AbstractFileSystem = kwargs.get("fs", self.fs)
+        with fs.open(input_file, errors=self.errors, encoding=self.encoding) as f:
+            # default mode is 'rb', we can cast the return value of f.read()
+            return cast(bytes, f.read())
+
     @staticmethod
     def load_file(
-        input_file: Path,
-        file_metadata: Callable[[str], Dict],
-        file_extractor: Dict[str, BaseReader],
+        input_file: Path | PurePosixPath,
+        file_metadata: Callable[[str], dict],
+        file_extractor: dict[str, BaseReader],
         filename_as_id: bool = False,
         encoding: str = "utf-8",
         errors: str = "ignore",
-        fs: Optional[fsspec.AbstractFileSystem] = None,
-    ) -> List[Document]:
-        """Static method for loading file.
+        raise_on_error: bool = False,
+        fs: fsspec.AbstractFileSystem | None = None,
+    ) -> list[Document]:
+        """
+        Static method for loading file.
 
         NOTE: necessarily as a static method for parallel processing.
 
@@ -350,6 +535,7 @@ class SimpleDirectoryReader(BaseReader):
             Default is utf-8.
         errors (str): how encoding and decoding errors are to be handled,
               see https://docs.python.org/3/library/functions.html#open
+        raise_on_error (bool): Whether to raise an error if a file cannot be read.
         fs (Optional[fsspec.AbstractFileSystem]): File system to use. Defaults
             to using the local file system. Can be changed to use any remote file system
 
@@ -359,8 +545,8 @@ class SimpleDirectoryReader(BaseReader):
         # TODO: make this less redundant
         default_file_reader_cls = SimpleDirectoryReader.supported_suffix_fn()
         default_file_reader_suffix = list(default_file_reader_cls.keys())
-        metadata: Optional[dict] = None
-        documents: List[Document] = []
+        metadata: dict | None = None
+        documents: list[Document] = []
 
         if file_metadata is not None:
             metadata = file_metadata(str(input_file))
@@ -376,7 +562,7 @@ class SimpleDirectoryReader(BaseReader):
 
             # load data -- catch all errors except for ImportError
             try:
-                kwargs = {"extra_info": metadata}
+                kwargs: dict[str, Any] = {"extra_info": metadata}
                 if fs and not is_default_fs(fs):
                     kwargs["fs"] = fs
                 docs = reader.load_data(input_file, **kwargs)
@@ -385,6 +571,8 @@ class SimpleDirectoryReader(BaseReader):
                 # about missing dependencies
                 raise ImportError(str(e))
             except Exception as e:
+                if raise_on_error:
+                    raise Exception("Error loading file") from e
                 # otherwise, just skip the file and report the error
                 print(
                     f"Failed to load file {input_file} with error: {e}. Skipping...",
@@ -402,9 +590,79 @@ class SimpleDirectoryReader(BaseReader):
             # do standard read
             fs = fs or get_default_fs()
             with fs.open(input_file, errors=errors, encoding=encoding) as f:
-                data = f.read().decode(encoding, errors=errors)
+                data = cast(bytes, f.read()).decode(encoding, errors=errors)
 
-            doc = Document(text=data, metadata=metadata or {})
+            doc = Document(text=data, metadata=metadata or {})  # type: ignore
+            if filename_as_id:
+                doc.id_ = str(input_file)
+
+            documents.append(doc)
+
+        return documents
+
+    @staticmethod
+    async def aload_file(
+        input_file: Path | PurePosixPath,
+        file_metadata: Callable[[str], dict],
+        file_extractor: dict[str, BaseReader],
+        filename_as_id: bool = False,
+        encoding: str = "utf-8",
+        errors: str = "ignore",
+        raise_on_error: bool = False,
+        fs: fsspec.AbstractFileSystem | None = None,
+    ) -> list[Document]:
+        """Load file asynchronously."""
+        # TODO: make this less redundant
+        default_file_reader_cls = SimpleDirectoryReader.supported_suffix_fn()
+        default_file_reader_suffix = list(default_file_reader_cls.keys())
+        metadata: dict | None = None
+        documents: list[Document] = []
+
+        if file_metadata is not None:
+            metadata = file_metadata(str(input_file))
+
+        file_suffix = input_file.suffix.lower()
+        if file_suffix in default_file_reader_suffix or file_suffix in file_extractor:
+            # use file readers
+            if file_suffix not in file_extractor:
+                # instantiate file reader if not already
+                reader_cls = default_file_reader_cls[file_suffix]
+                file_extractor[file_suffix] = reader_cls()
+            reader = file_extractor[file_suffix]
+
+            # load data -- catch all errors except for ImportError
+            try:
+                kwargs: dict[str, Any] = {"extra_info": metadata}
+                if fs and not is_default_fs(fs):
+                    kwargs["fs"] = fs
+                docs = await reader.aload_data(input_file, **kwargs)
+            except ImportError as e:
+                # ensure that ImportError is raised so user knows
+                # about missing dependencies
+                raise ImportError(str(e))
+            except Exception as e:
+                if raise_on_error:
+                    raise
+                # otherwise, just skip the file and report the error
+                print(
+                    f"Failed to load file {input_file} with error: {e}. Skipping...",
+                    flush=True,
+                )
+                return []
+
+            # iterate over docs if needed
+            if filename_as_id:
+                for i, doc in enumerate(docs):
+                    doc.id_ = f"{input_file!s}_part_{i}"
+
+            documents.extend(docs)
+        else:
+            # do standard read
+            fs = fs or get_default_fs()
+            with fs.open(input_file, errors=errors, encoding=encoding) as f:
+                data = cast(bytes, f.read()).decode(encoding, errors=errors)
+
+            doc = Document(text=data, metadata=metadata or {})  # type: ignore
             if filename_as_id:
                 doc.id_ = str(input_file)
 
@@ -415,10 +673,11 @@ class SimpleDirectoryReader(BaseReader):
     def load_data(
         self,
         show_progress: bool = False,
-        num_workers: Optional[int] = None,
-        fs: Optional[fsspec.AbstractFileSystem] = None,
-    ) -> List[Document]:
-        """Load data from the input directory.
+        num_workers: int | None = None,
+        fs: fsspec.AbstractFileSystem | None = None,
+    ) -> list[Document]:
+        """
+        Load data from the input directory.
 
         Args:
             show_progress (bool): Whether to show tqdm progress bars. Defaults to False.
@@ -435,11 +694,14 @@ class SimpleDirectoryReader(BaseReader):
         fs = fs or self.fs
 
         if num_workers and num_workers > 1:
-            if num_workers > multiprocessing.cpu_count():
+            num_cpus = multiprocessing.cpu_count()
+            if num_workers > num_cpus:
                 warnings.warn(
                     "Specified num_workers exceed number of CPUs in the system. "
                     "Setting `num_workers` down to the maximum CPU count."
                 )
+                num_workers = num_cpus
+
             with multiprocessing.get_context("spawn").Pool(num_workers) as p:
                 results = p.starmap(
                     SimpleDirectoryReader.load_file,
@@ -450,6 +712,7 @@ class SimpleDirectoryReader(BaseReader):
                         repeat(self.filename_as_id),
                         repeat(self.encoding),
                         repeat(self.errors),
+                        repeat(self.raise_on_error),
                         repeat(fs),
                     ),
                 )
@@ -469,16 +732,66 @@ class SimpleDirectoryReader(BaseReader):
                         filename_as_id=self.filename_as_id,
                         encoding=self.encoding,
                         errors=self.errors,
+                        raise_on_error=self.raise_on_error,
                         fs=fs,
                     )
                 )
 
         return self._exclude_metadata(documents)
 
+    async def aload_data(
+        self,
+        show_progress: bool = False,
+        num_workers: int | None = None,
+        fs: fsspec.AbstractFileSystem | None = None,
+    ) -> list[Document]:
+        """
+        Load data from the input directory.
+
+        Args:
+            show_progress (bool): Whether to show tqdm progress bars. Defaults to False.
+            num_workers  (Optional[int]): Number of workers to parallelize data-loading over.
+            fs (Optional[fsspec.AbstractFileSystem]): File system to use. If fs was specified
+                in the constructor, it will override the fs parameter here.
+
+        Returns:
+            List[Document]: A list of documents.
+        """
+        files_to_process = self.input_files
+        fs = fs or self.fs
+
+        coroutines = [
+            SimpleDirectoryReader.aload_file(
+                input_file,
+                self.file_metadata,
+                self.file_extractor,
+                self.filename_as_id,
+                self.encoding,
+                self.errors,
+                self.raise_on_error,
+                fs,
+            )
+            for input_file in files_to_process
+        ]
+
+        if num_workers:
+            document_lists = await run_jobs(
+                coroutines, show_progress=show_progress, workers=num_workers
+            )
+        elif show_progress:
+            _asyncio = get_asyncio_module(show_progress=show_progress)
+            document_lists = await _asyncio.gather(*coroutines)
+        else:
+            document_lists = await asyncio.gather(*coroutines)
+        documents = [doc for doc_list in document_lists for doc in doc_list]
+
+        return self._exclude_metadata(documents)
+
     def iter_data(
         self, show_progress: bool = False
-    ) -> Generator[List[Document], Any, Any]:
-        """Load data iteratively from the input directory.
+    ) -> Generator[list[Document], Any, Any]:
+        """
+        Load data iteratively from the input directory.
 
         Args:
             show_progress (bool): Whether to show tqdm progress bars. Defaults to False.
@@ -499,6 +812,7 @@ class SimpleDirectoryReader(BaseReader):
                 filename_as_id=self.filename_as_id,
                 encoding=self.encoding,
                 errors=self.errors,
+                raise_on_error=self.raise_on_error,
                 fs=self.fs,
             )
 
